/********************************************************
* Author  : Yonghyun Hwang
* Date    : 2008. 02. 05
* Modified By  : Yonghyun Hwang
* Last Update : 2008. 10. 08
* Description : IPC
********************************************************/
#ifndef IPC_H
#define IPC_H

#include "rtos.h"

namespace ese_rtos {

#define DEFINE_RTOS_IPC_DH(rtos_name, ch_name, send_ipc_name, recv_ipc_name) \
  ESE_CH_DH * rtos_name ## _ ## ch_name ;				\
  extern "C" void send_ipc_name (void * ptr, int size) {		\
    rtos_name ## _ ## ch_name ->send(ptr, (unsigned int) size);         \
  }                                                                     \
  									\
  extern "C" void recv_ipc_name (void * ptr, int size) {		\
    rtos_name ## _ ## ch_name ->recv(ptr, (unsigned int) size);         \
  }

// b_write and b_read decides whether they are blocking or not
// if they are blocking read or write, 'b_read' and 'b_write' is passed to the MACRO as its argument
// if not, just 'read' and 'write' is passed
#define DEFINE_RTOS_IPC_FIFO(rtos_name, ch_name, b_write_ipc_name, write_ipc_name, b_read_ipc_name, read_ipc_name) \
  ESE_CH_FIFO * rtos_name ## _ ## ch_name ;				\
  extern "C" void b_write_ipc_name (void * ptr, int size) {             \
    rtos_name ## _ ## ch_name -> b_write (ptr, (unsigned int) size);    \
  }                                                                     \
                                                                        \
  extern "C" void write_ipc_name (void * ptr, int size) {               \
    rtos_name ## _ ## ch_name -> nb_write (ptr, (unsigned int) size);      \
  }                                                                     \
  extern "C" void b_read_ipc_name (void * ptr, int size) {              \
    rtos_name ## _ ## ch_name -> b_read (ptr, (unsigned int) size);     \
  }                                                                     \
                                                                        \
  extern "C" void read_ipc_name (void * ptr, int size) {                \
    rtos_name ## _ ## ch_name -> nb_read (ptr, (unsigned int) size);       \
  }

  class i_ESE_CH_DH_IN : virtual public sc_interface {

  public:
    virtual unsigned int send(void * ptr, unsigned int size) = 0;
  };

  class i_ESE_CH_DH_OUT : virtual public sc_interface {

  public:
    virtual unsigned int recv(void * ptr, unsigned int size) = 0;
  };

  class i_ESE_CH_FIFO_IN : virtual public sc_interface {

  public:
    virtual bool nb_write(const void * ptr, unsigned int size) = 0;
    virtual bool b_write(const void * ptr, unsigned int size) = 0;
  };

  class i_ESE_CH_FIFO_OUT : virtual public sc_interface {

  public:
    virtual bool nb_read(void * ptr, unsigned int size) = 0;
    virtual bool b_read(void * ptr, unsigned int size) = 0;
  };

  class ESE_CH_DH : public sc_module, public i_ESE_CH_DH_IN, public i_ESE_CH_DH_OUT {
    
  public:
    ESE_CH_DH(sc_module_name name, unsigned int srcPid, unsigned int destPid);
    ~ESE_CH_DH() {}

    static ESE_CH_DH * create(ESE_RTOS * rtos, std::string chName, unsigned int srcPid, unsigned int destPid);
    
    /** it simply copies ptr into buf */
    unsigned int send(void * ptr, unsigned int size);
    /** when calling recv, data is copied into the mem location by ptr */
    unsigned int recv(void * ptr, unsigned int size);


  public:
    sc_port<i_ESE_RTOS> rtos_call;

  private:

    ESE_RTOS * rtos;

    /** because we don't have 'port' to thread, srcPid and destPid should be kept
	this is temporary implementation and need to be replaced w/ sc_port */
    unsigned int srcPid, destPid;

    /** if buf is equal to NULL, it implies data in buf is invalid */
    void * buf; 
    unsigned int size;


  };

  /** Implementation of RTOS FIFO */
  class ESE_CH_FIFO : public sc_module, public i_ESE_CH_FIFO_IN, public i_ESE_CH_FIFO_OUT {

#define DEFAULT_QUEUE_SIZE 1024

  public:
    ESE_CH_FIFO(sc_module_name name, unsigned int srcPid, unsigned int destPid, unsigned int qSize = 0);
    ~ESE_CH_FIFO();

    static ESE_CH_FIFO * create(ESE_RTOS * rtos, std::string chName, unsigned int srcPid, unsigned int destPid, unsigned int qSize = 0);

    bool nb_write(const void * src, unsigned int size);
    bool nb_read(void * dest, unsigned int size);

    bool b_write(const void * src, unsigned int size);
    bool b_read(void * dest, unsigned int size);

    void clear(void);
    void do_check(void);

  private:
    const bool is_empty(void) { return (front == rear); }
    const bool is_full(void) { return (rear == (front + 1) % bufSize); }
    const unsigned int size_of_free_space(void);
    const unsigned int size_of_used_space(void);

  private:

    ESE_RTOS * rtos;

    /** because we don't have 'port' to thread, srcPid and destPid should be kept
	this is temporary implementation and need to be replaced w/ sc_port */
    unsigned int srcPid, destPid;

    char * buf;
    unsigned int bufSize;
    unsigned int front, rear;

    sc_port<i_ESE_RTOS> rtos_call;
  };

} // end of namespace ese_rtos

#endif // IPC_H
