#ifndef RTOS_H
#define RTOS_H

#include <iostream>
#include <string>
#include <vector>
#include <queue>

#include "systemc.h"
#include "rtos_thread.h"
#include "rtos_queue.h"
#include "rtos_stats.h"

/** Maximum number of threads supported by current ESE RTOS implementation */
#define MAX_NUM_OF_THREAD 128
#define INTR_PRIORITY 0

using namespace sc_core;



namespace ese_rtos {

  class ESE_Thread;
  class ESE_RTOS;
  class RTOS_Stats;

  typedef struct RTOS_Conf {
    std::string rtosType, schedPolicy;
    unsigned int timeSlice;

    //unsigned int contextSwitchLat, semPLat, semVLat;
    //unsigned int sharedMemReadLat, sharedMemWriteLat;
  } RTOS_Conf;


  class PtrToTCB {
  public:
  PtrToTCB() : pid(0), priority(0) {}
    ~PtrToTCB() {}

    void prt_content(void) {
      std::cout << "pid(" << pid << ")::prior(" << priority << ")\n";
    }

  public:
    unsigned int pid;
    unsigned int priority;
  };

  inline bool operator<(const PtrToTCB & lhs, const PtrToTCB & rhs) {
    return lhs.priority < rhs.priority;
  }

  class RT_Event {
  public:
    RT_Event() {}
    ~RT_Event() {}

  public:
    std::queue<unsigned int> waitQueue; // store pid only
    std::queue<ESE_RTOS *> ptrQueue;
  };

  /*
    class RT_Mutex {
    public:
    RT_Mutex() : flag(false) {}
    ~RT_Mutex() {}

    public:
    bool flag;
    std::queue<unsigned int> waitQueue;
    std::queue<ESE_RTOS *> ptrQueue;
    };
  */

  /** Declare interface common for all RTOS implementations */
  class i_ESE_RTOS : virtual public sc_interface {

  public:

    /**  Invoke first thread to start multitasking in the RTOS
	 Assumption! This should be called after all the threads are created! */
    virtual void start(void) = 0;

    /** Task initialization
	Once systemc thread gets created, it needs to wait scheduler to control the flow of thread */
    virtual void task_init(unsigned int pid) = 0;

    /** Task creation
	Task Control Block for a systemc thread is created into a table for TCB.
	This function returns pid. If 0 is returned, it means it fails to create a thread */
    virtual unsigned int task_create(sc_module_name name, unsigned int priority, int (*_func)(void), void * param) = 0;

    /** Wait for activation by other thread */
    virtual void task_wait(void) = 0;

    /** Activate waiting thread
	If the thread is already running or in the ready queue,
	nothing happens */
    virtual void task_activate(unsigned int pid) = 0;

    /** Task termination
	Task completes its mission, it should notify rtos that it will be terminated */
    virtual void task_terminate(void) = 0;

    /** Advance system clock by delay
	This results in scheduling in TTLM
	Please note that the delay is in the number of cycles */
    virtual void time_wait(unsigned int delay) = 0;


    /** Task sleep (now empty)
	Make task sleep and wake up after a given delay
	Please not that the delay is in the number of millisecond */
    virtual void task_sleep(unsigned int mSecs) = 0;

    virtual void event_wait(RT_Event & event) = 0;
    virtual void event_notify(RT_Event & event) = 0;

    /** Pre interrupt wait
	Make task pended and run scheduling without actual context switch.
	For this, status for the task becomes 'pending' and 
	other task gets ready to run based on scheduling scheme
	Once this is called, conceptually, 2 tasks are running at the same time on the ESE RTOS in systemC kernel
	The task calling pre_int_wait will be suspended in the ubc call
	Please note that system clock for a PE should be advanced in the ubc even when the PE is in idle status
     */
    virtual unsigned int pre_int_wait(void) = 0;

    /** Post interrupt wait 
	Find and set the current TCB based pid
     */
    virtual void post_int_wait(unsigned int pid) = 0;

    /**  Interrupt enter
	 In side of the ubc call, the task is woke up. Task wakeup means that the PE having the task gets an interrupt. To emulate interrupt handling in the PE,
	 this function is called. The function adds TCB for the task into task ready queue with highest priority in ESE RTOS. If the PE was idle, it lets the task continue with highest priority
	 ASSUMPTION! no threads in ESE RTOS have '0' priority
     */
    virtual void isr_enter(void) = 0;

    /** Interrupt delay (empty for now)
	Atomically advance system clock
	Delay is in the number of cycles */
    virtual void isr_delay(void) = 0;

    /** Interrrupt return 
	After emulating interrupt handling in int_enter() and int_delay(),
	system mode should get back to the user mode from the kernel mode.
	int_return emulates this. The function changes a status for the task
	into 'ready' state and adds TCB to the ready queue for next scheduling
     */
    virtual void isr_return(void) = 0;

    /** Switch to other task */
    virtual void run_sched(void) = 0;

    virtual void acquire(void) = 0;
    virtual void release(void) = 0;

    virtual ESE_Thread * get_tcb(unsigned int pid) = 0;

    //virtual void acquire(RT_Mutex & _mutex) = 0;
    //virtual void release(RT_Mutex & _mutex) = 0;

    //virtual class Ipc * ipc_create(const std::string & ipcType, unsigned int srcPid, unsigned int destPid, unsigned int size) = 0;

    virtual void prt_sc_time(void) = 0;

    virtual unsigned int get_sys_time(void) = 0;
  };

  
  class ESE_RTOS : public sc_module, public i_ESE_RTOS {

    //friend class Ipc;

  public:
    ESE_RTOS(sc_module_name name, const RTOS_Conf & conf );
    ~ESE_RTOS();

  public:

    void start(void);
    void task_init(unsigned int pid);

    unsigned int task_create(sc_module_name name, unsigned int priority, int (*_func)(void), void * param);

    void run_sched(void);

    void task_wait(void);
    void task_activate(unsigned int pid);

    void task_terminate(void);

    void task_sleep(unsigned int mSecs);

    unsigned int pre_int_wait(void);
    void post_int_wait(unsigned int pid);
    void isr_enter(void);
    void isr_delay(void);
    void isr_return(void);

    void event_wait(RT_Event & event);
    void event_notify(RT_Event & event);

    void acquire(void);
    void release(void);

    ESE_Thread * get_tcb(unsigned int pid);

    //void acquire(RT_Mutex & _mutex);
    //void release(RT_Mutex & _mutex);

    //class Ipc * ipc_create(const std::string & ipcType, unsigned int srcPid, unsigned int destPid, unsigned int size);

    /** advance system clock
	just wrap up code for sc_core::wait */
    void adv_sys_clock(unsigned int delay);


    void time_wait(unsigned int delay);

    void prt_sc_time(void);

    unsigned long long int get_rtos_idle_time(void);
    unsigned long long int get_rtos_overhead(void);

    unsigned int get_sys_time(void);

    void dump_rtos_statistics(void);

    /** dump rtos status to faciliate debugging
	list all the threads' status in ESE RTOS
    */
    void dbg_dump_rtos_status(void);
    void prt_curr_sc_simcontext(void);


  private:
    /** Task dispatcher.
	Please note that scheduling policy is implemented in queue handling routine.
	The dispatcher simply pop up the first element in the queue and make it runnable */
    void task_dispatcher(void);
    void cpu_idle(void);
    void make_task_ready(ESE_Thread * _thread);

    /** make ESE thread wait for next scheduling */
    inline void wait_for_sched(unsigned int pid); 
    /** make ESE thread wake from ready status */
    inline void send_sched(unsigned int pid);  

  public:

    /** collecting rtos statistics */
    class RTOS_Stats rtosStats;

    ESE_Thread * curTask;

  private:

    /** Table stores Task Control Blocks */
    ESE_Thread * tcbTable[MAX_NUM_OF_THREAD];
    unsigned int tcbTabSize;

    /** Every thread has an systemc event for scheduling */
    sc_event * schedEvtTable[MAX_NUM_OF_THREAD];
    unsigned int schedEvtTabSize;


    /** Two variable below are used to check when all threads in rtos get complete */
    unsigned int numOfDoneThreads;
    sc_event threadDone;

    /** Name for rtos */
    std::string rtosName;
    /** Type for rtos */
    std::string rtosType;

    /** queue implements scheduling policy
	in side of queue, there exists only pointer to the table for TCB */
    TaskQueue<PtrToTCB> * readyQueue;
    TaskQueue<PtrToTCB> * mutexQueue;

    bool mutexLocked;

    /** time slice.
	in case of RR sched, the variable is used for time slot allocation.
	in case of Prior sched, it is used to increase system clock when all the threads are waiting for external event to come. */
    unsigned int timeSlice;
    /** systemc event to wake up idle rtos when external intr is coming up */
    sc_event intrEvt;

    bool isRoundRobin; // this is temporary variable. should be removed somehow in the future

    // this variable is used when doing scheduling
    // in ESE RTOS, two sysc threads can be run internally
    // if one of them is back under control of RTOS, that need to be notified
    // unless, priority scheduler doesn't know whether higher prior task is back to the ready queue.
    // refer to time_wait function.
    // NOTICE: it's solely for sim speed
    // below var only get true in task_sleep and post_int_wait
    // time_wait set false
    bool existHighPrior;
  };
} // end of namespace ese_rtos

extern "C" void dump_rtos_status(void * ptrRTOS);

#endif // RTOS_H
