/*
 * libmad - MPEG audio decoder library
 * Copyright (C) 2000-2004 Underbit Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: synth.c,v 1.1 2008/11/05 17:41:29 yjahn Exp $
 */

# ifdef HAVE_CONFIG_H
#  include "config.h"
# endif

# include "global.h"

# include "fixed.h"
# include "frame.h"
# include "synth.h"

/*
 * NAME:	synth->init()
 * DESCRIPTION:	initialize synth struct
 */
void mad_synth_init(struct mad_synth *synth)
{
  mad_synth_mute(synth);

  synth->phase = 0;

  synth->pcm.samplerate = 0;
  synth->pcm.channels   = 0;
  synth->pcm.length     = 0;
}

/*
 * NAME:	synth->mute()
 * DESCRIPTION:	zero all polyphase filterbank values, resetting synthesis
 */
void mad_synth_mute(struct mad_synth *synth)
{
  unsigned int ch, s, v;
  for (ch = 0; ch < 2; ++ch) {
    for (s = 0; s < 16; ++s) {
      for (v = 0; v < 8; ++v) {
	synth->filter[ch][0][0][s][v] = synth->filter[ch][0][1][s][v] =
	synth->filter[ch][1][0][s][v] = synth->filter[ch][1][1][s][v] = 0;
      }
    }
  }
}

/*
 * An optional optimization called here the Subband Synthesis Optimization
 * (SSO) improves the performance of subband synthesis at the expense of
 * accuracy.
 *
 * The idea is to simplify 32x32->64-bit multiplication to 32x32->32 such
 * that extra scaling and rounding are not necessary. This often allows the
 * compiler to use faster 32-bit multiply-accumulate instructions instead of
 * explicit 64-bit multiply, shift, and add instructions.
 *
 * SSO works like this: a full 32x32->64-bit multiply of two mad_fixed_t
 * values requires the result to be right-shifted 28 bits to be properly
 * scaled to the same fixed-point format. Right shifts can be applied at any
 * time to either operand or to the result, so the optimization involves
 * careful placement of these shifts to minimize the loss of accuracy.
 *
 * First, a 14-bit shift is applied with rounding at compile-time to the D[]
 * table of coefficients for the subband synthesis window. This only loses 2
 * bits of accuracy because the lower 12 bits are always zero. A second
 * 12-bit shift occurs after the DCT calculation. This loses 12 bits of
 * accuracy. Finally, a third 2-bit shift occurs just before the sample is
 * saved in the PCM buffer. 14 + 12 + 2 == 28 bits.
 */

/* FPM_DEFAULT without OPT_SSO will actually lose accuracy and performance */

# if defined(FPM_DEFAULT) && !defined(OPT_SSO)
#  define OPT_SSO
# endif

/* second SSO shift, with rounding */

# if defined(OPT_SSO)
#  define SHIFT(x)  (((x) + (1L << 11)) >> 12)
# else
#  define SHIFT(x)  (x)
# endif

/* possible DCT speed optimization */

# if defined(OPT_SPEED) && defined(MAD_F_MLX)
#  define OPT_DCTO
#  define MUL(x, y)  \
    ({ mad_fixed64hi_t hi;  \
       mad_fixed64lo_t lo;  \
       MAD_F_MLX(hi, lo, (x), (y));  \
       hi << (32 - MAD_F_SCALEBITS - 3);  \
    })
# else
#  undef OPT_DCTO
//#  define MUL(x, y)  mad_f_mul((x), (y))
#  define MUL(x, y)  ((((x) + (1L << 11)) >> 12) *  \
                                 (((y) + (1L << 15)) >> 16))
# endif


# undef MUL
# undef SHIFT

/* third SSO shift and/or D[] optimization preshift */

# if defined(OPT_SSO)
#  if MAD_F_FRACBITS != 28
#   error "MAD_F_FRACBITS must be 28 to use OPT_SSO"
#  endif
#  define ML0(hi, lo, x, y)	((lo)  = (x) * (y))
#  define MLA(hi, lo, x, y)	((lo) += (x) * (y))
#  define MLN(hi, lo)		((lo)  = -(lo))
#  define MLZ(hi, lo)		((void) (hi), (mad_fixed_t) (lo))
#  define SHIFT(x)		((x) >> 2)
#  define PRESHIFT(x)		((MAD_F(x) + (1L << 13)) >> 14)
# else
#  define ML0(hi, lo, x, y)	MAD_F_ML0((hi), (lo), (x), (y))
#  define MLA(hi, lo, x, y)	MAD_F_MLA((hi), (lo), (x), (y))
#  define MLN(hi, lo)		MAD_F_MLN((hi), (lo))
#  define MLZ(hi, lo)		MAD_F_MLZ((hi), (lo))
#  define SHIFT(x)		(x)
#  if defined(MAD_F_SCALEBITS)
#   undef  MAD_F_SCALEBITS
#   define MAD_F_SCALEBITS	(MAD_F_FRACBITS - 12)
#   define PRESHIFT(x)		(MAD_F(x) >> 12)
#  else
#   define PRESHIFT(x)		MAD_F(x)
#  endif
# endif

mad_fixed_t const D[17][32] = {
# include "D.dat"
};

/*
 * The following utility routine performs simple rounding, clipping, and
 * scaling of MAD's high-resolution samples down to 16 bits. It does not
 * perform any dithering or noise shaping, which would be recommended to
 * obtain any exceptional audio quality. It is therefore not recommended to
 * use this routine if high-quality output is desired.
 */

static /*inline*/ 
signed int scale(mad_fixed_t sample)
{
  /* round */
  sample += (1L << (MAD_F_FRACBITS - 16));

  /* clip */
  if (sample >= MAD_F_ONE)
    sample = MAD_F_ONE - 1;
  else if (sample < -MAD_F_ONE)
    sample = -MAD_F_ONE;

  /* quantize */
  return sample >> (MAD_F_FRACBITS + 1 - 16);
}


mad_fixed_t* calc_sample (
		  mad_fixed_t (*filter)[2][2][16][8],
		  unsigned int phase,
		  mad_fixed_t *pcm1) {
  unsigned int sb, pe, po;
  mad_fixed_t *pcm2;
  register mad_fixed_t (*fe)[8], (*fx)[8], (*fo)[8];
  register mad_fixed_t const (*Dptr)[32], *ptr;
  register mad_fixed64hi_t hi;
  register mad_fixed64lo_t lo;

  pe = phase & ~1;
  po = ((phase - 1) & 0xf) | 1;

  /* calculate 32 samples */

  fe = &(*filter)[0][ phase & 1][0];
  fx = &(*filter)[0][~phase & 1][0];
  fo = &(*filter)[1][~phase & 1][0];

  Dptr = &D[0];

  ptr = *Dptr + po;
  ML0(hi, lo, (*fx)[0], ptr[ 0]);
  MLA(hi, lo, (*fx)[1], ptr[14]);
  MLA(hi, lo, (*fx)[2], ptr[12]);
  MLA(hi, lo, (*fx)[3], ptr[10]);
  MLA(hi, lo, (*fx)[4], ptr[ 8]);
  MLA(hi, lo, (*fx)[5], ptr[ 6]);
  MLA(hi, lo, (*fx)[6], ptr[ 4]);
  MLA(hi, lo, (*fx)[7], ptr[ 2]);
  MLN(hi, lo);

  ptr = *Dptr + pe;
  MLA(hi, lo, (*fe)[0], ptr[ 0]);
  MLA(hi, lo, (*fe)[1], ptr[14]);
  MLA(hi, lo, (*fe)[2], ptr[12]);
  MLA(hi, lo, (*fe)[3], ptr[10]);
  MLA(hi, lo, (*fe)[4], ptr[ 8]);
  MLA(hi, lo, (*fe)[5], ptr[ 6]);
  MLA(hi, lo, (*fe)[6], ptr[ 4]);
  MLA(hi, lo, (*fe)[7], ptr[ 2]);

  *pcm1 = SHIFT(MLZ(hi, lo));
  *pcm1 = scale(*pcm1);
  pcm1++;

  pcm2 = pcm1 + 30;

  for (sb = 1; sb < 16; ++sb) {
    ++fe;
    ++Dptr;

    /* D[32 - sb][i] == -D[sb][31 - i] */

    ptr = *Dptr + po;
    ML0(hi, lo, (*fo)[0], ptr[ 0]);
    MLA(hi, lo, (*fo)[1], ptr[14]);
    MLA(hi, lo, (*fo)[2], ptr[12]);
    MLA(hi, lo, (*fo)[3], ptr[10]);
    MLA(hi, lo, (*fo)[4], ptr[ 8]);
    MLA(hi, lo, (*fo)[5], ptr[ 6]);
    MLA(hi, lo, (*fo)[6], ptr[ 4]);
    MLA(hi, lo, (*fo)[7], ptr[ 2]);
    MLN(hi, lo);

    ptr = *Dptr + pe;
    MLA(hi, lo, (*fe)[7], ptr[ 2]);
    MLA(hi, lo, (*fe)[6], ptr[ 4]);
    MLA(hi, lo, (*fe)[5], ptr[ 6]);
    MLA(hi, lo, (*fe)[4], ptr[ 8]);
    MLA(hi, lo, (*fe)[3], ptr[10]);
    MLA(hi, lo, (*fe)[2], ptr[12]);
    MLA(hi, lo, (*fe)[1], ptr[14]);
    MLA(hi, lo, (*fe)[0], ptr[ 0]);
    *pcm1 = SHIFT(MLZ(hi, lo));
    *pcm1 = scale(*pcm1);
    pcm1++;


    ptr = *Dptr - pe;
    ML0(hi, lo, (*fe)[0], ptr[31 - 16]);
    MLA(hi, lo, (*fe)[1], ptr[31 - 14]);
    MLA(hi, lo, (*fe)[2], ptr[31 - 12]);
    MLA(hi, lo, (*fe)[3], ptr[31 - 10]);
    MLA(hi, lo, (*fe)[4], ptr[31 -  8]);
    MLA(hi, lo, (*fe)[5], ptr[31 -  6]);
    MLA(hi, lo, (*fe)[6], ptr[31 -  4]);
    MLA(hi, lo, (*fe)[7], ptr[31 -  2]);

    ptr = *Dptr - po;
    MLA(hi, lo, (*fo)[7], ptr[31 -  2]);
    MLA(hi, lo, (*fo)[6], ptr[31 -  4]);
    MLA(hi, lo, (*fo)[5], ptr[31 -  6]);
    MLA(hi, lo, (*fo)[4], ptr[31 -  8]);
    MLA(hi, lo, (*fo)[3], ptr[31 - 10]);
    MLA(hi, lo, (*fo)[2], ptr[31 - 12]);
    MLA(hi, lo, (*fo)[1], ptr[31 - 14]);
    MLA(hi, lo, (*fo)[0], ptr[31 - 16]);
    *pcm2 = SHIFT(MLZ(hi, lo));
    *pcm2 = scale(*pcm2);
    pcm2--;


    ++fo;
  }

  ++Dptr;

  ptr = *Dptr + po;
  ML0(hi, lo, (*fo)[0], ptr[ 0]);
  MLA(hi, lo, (*fo)[1], ptr[14]);
  MLA(hi, lo, (*fo)[2], ptr[12]);
  MLA(hi, lo, (*fo)[3], ptr[10]);
  MLA(hi, lo, (*fo)[4], ptr[ 8]);
  MLA(hi, lo, (*fo)[5], ptr[ 6]);
  MLA(hi, lo, (*fo)[6], ptr[ 4]);
  MLA(hi, lo, (*fo)[7], ptr[ 2]);
  *pcm1 = SHIFT(-MLZ(hi, lo));
  *pcm1 = scale(*pcm1);

  pcm1 += 16;
  return pcm1;
}

